from typing import Dict, Callable
from pathlib import Path
import os
import subprocess
from tkinter import messagebox


import connection_strings

_REQUIRED_ATTRS = ["user", "password", "db_type", "dir", "version", "disable"]


def is_cellario_running() -> bool:
    return bool(len([x for x in subprocess.check_output(['tasklist']).split(b'\r\n') if b"Cellario.exe" in x]))


class Instance:
    def __init__(self, name: str, input_dict: Dict):
        self.name = name
        self.disable = False
        for (key, value) in input_dict.items():
            match key.lower():
                case "databaseuser" | "user":
                    self.user = value
                case "databasepassword" | "password":
                    self.password = value
                case "databasetype" | "type":
                    self.db_type = value
                case "cellariodirectory" | "directory" | "dir":
                    self.dir = value
                case "version":
                    self.version = value
                case "disable":
                    if value.lower() in ("true", "1", "y", "yes"):
                        self.disable = True
                case _:
                    print(f"Unexpected key: {key}, dropping this value.")
        if not self._validate():
            raise ValueError("Instance missing one or more valid keys.")

    def version_is_43_or_higher(self) -> bool:
        version_numbers = self.version.split('.')
        if int(version_numbers[0]) < 4:
            return False
        if int(version_numbers[1]) >= 3:
            return True
        return False

    def _validate(self) -> bool:
        for attr in _REQUIRED_ATTRS:
            if not hasattr(self, attr):
                return False
        return True

    def __repr__(self):
        attrs = ', '.join(f'{key}={value!r}' for key,
                          value in self.__dict__.items())
        return f'{self.__class__.__name__}({attrs})'


def set_symlink_for_instance(instance: Instance):
    target_path = r"C:\Program Files\HighRes Biosolutions\Cellario"
    if Path(target_path).is_dir() and not Path(target_path).is_symlink():
        print("Real CellarioScheduler directory already exists; it will be moved")
        import shutil
        import random
        import string

        hrb_dir = r"C:\Program Files\HighRes Biosolutions"
        new_dir_name = "Cellario_old_" + \
            ''.join(random.choices(string.ascii_lowercase, k=8))
        shutil.move(target_path, hrb_dir + "\\" + new_dir_name)
    if Path(target_path).is_dir() and Path(target_path).is_symlink():
        os.unlink(target_path)

    os.symlink(instance.dir, target_path, target_is_directory=True)


def gen_function_for_instance(instance: Instance) -> Callable:
    match instance.db_type:
        case "postgres":
            conn_string = connection_strings.generate_postgres_conn_string(
                instance)
        case "oracle":
            conn_string = connection_strings.generate_oracle_conn_string(
                instance)
        case _:
            # What???
            raise ValueError()

    def retval():
        if not is_cellario_running():
            # Run symlink first, makes editing json easier in 4.3+ case
            set_symlink_for_instance(instance)
            connection_strings.set_conn_string(
                conn_string,
                instance.db_type,
                instance.version_is_43_or_higher())
            subprocess.Popen([r"C:\Program Files\HighRes Biosolutions\Cellario\Cellario.exe"],
                             cwd=r"C:\Program Files\HighRes Biosolutions\Cellario")
            # exit(0)
        else:
            messagebox.showerror(
                title="Cellario already running", message="Cellario is already running!")
    return retval
